(*i $Id: gui.ml,v 1.3 2004/06/07 07:44:49 conchon Exp $ i*)

open Types
open Init

let _ = GMain.Main.init ()

let window = 
  GWindow.window 
    ~width:(Gdk.Screen.width ()) ~height:(Gdk.Screen.height () - 30) ()

let vbox = GPack.vbox ~packing:window#add ()

let font = 
  Gdk.Font.load_fontset 
    (*"-adobe-courier-medium-r-normal--12-120-75-75-m-70-iso8859-1"*)
    "-adobe-courier-medium-r-normal--10-100-75-75-m-60-iso8859-15"
(*    "-adobe-courier-medium-r-normal--10-100-75-75-m-60-iso8859-15"*)

let default_size = 20
let size = ref default_size

let see_markers = ref true

let hexsize = 
  let ratio = sqrt 11. /. 4. -. 1. /. 3. in
  fun size -> int_of_float (ratio *. float size)

let hexdim size = 
  let h = hexsize size / 2 in 
  let m = size / 2 in
  m, m - h, m + h

let hexsides ~s ~m ~h1 ~h2 =
  let h = [| 0, h1; m, 0; s, h1; s, h2; m, s; 0, h2 |] in
  fun ~x ~y -> Array.map (fun (i, j) -> i + x, y + j) h

(* Return the graphical position from a model position [(x, y)]. *)
let real_pos s m h2 x y = x * s + y mod 2 * m, y * h2
    
let draw_ant (pixmap:GDraw.pixmap) x y ant =
  let m, h1, h2 = hexdim !size in
  let h1 = h1 * 10 / 9 in
  let h2 = h2 * 9 / 10 in
  (match ant.color, ant.is_alive with
     | Red, true -> pixmap#set_foreground (`NAME "brown")
     | Red, false -> pixmap#set_foreground (`NAME "pink")
     | Black, true -> pixmap#set_foreground `BLACK
     | Black, false -> pixmap#set_foreground `WHITE);
  let triangle =
    match ant.direction with
      | 0 -> [ h1, h1; h2, m; h1, h2 ]
      | 1 -> [ h2, h1; h2, h2; h1, h2 ]
      | 2 -> [ h1, h1; h2, h2; h1, h2 ]
      | 3 -> [ h1, m; h2, h1; h2, h2 ]
      | 4 -> [ h1, h1; h2, h1; h1, h2 ]
      | 5 -> [ h1, h1; h2, h1; h2, h2 ]
      | _ -> assert false
  in
  pixmap#polygon 
    ~filled:true (List.map (fun (a, b) -> a + x, b + y) triangle);
  pixmap#set_foreground (`NAME "green");
  if ant.has_food then pixmap#string "f" ~x:(x+m-5) ~y:(y+m) ~font

let check_any_marker m =
  m.(0) || m.(1) || m.(2) || m.(3) || m.(4) || m.(5)

let draw_hex (pixmap:GDraw.pixmap) border hex =
  assert (Array.length border = 6);
  pixmap#set_foreground `WHITE;
  let l = Array.to_list border in
  pixmap#polygon ~filled:true l;
  if hex.rocky then begin
    pixmap#set_foreground (`NAME "green");
    pixmap#polygon ~filled:true l;
  end else begin
    let x1, y1 = border.(1) in
    let x2, y2 = border.(5) in
    if !see_markers then begin
      if check_any_marker hex.black_markers then begin
	pixmap#set_foreground (`NAME "gray");
	pixmap#polygon
	  ~filled:true [ border.(1); border.(2); border.(3); border.(4) ]
      end;
      if hex.red_markers.(5) then pixmap#set_foreground (`NAME "darkblue")
      else if hex.red_markers.(4) then pixmap#set_foreground (`NAME "blue")
      else if hex.red_markers.(3) then pixmap#set_foreground (`NAME "cyan")
      else if hex.red_markers.(2) then pixmap#set_foreground (`NAME "orange")
      else if hex.red_markers.(1) then pixmap#set_foreground (`NAME "pink")
      else if hex.red_markers.(0) then pixmap#set_foreground (`NAME "yellow")
      else pixmap#set_foreground `WHITE;
      pixmap#polygon
	~filled:true [ border.(1); border.(4); border.(5); border.(0) ]
    end;
    Gaux.may (fun n -> draw_ant pixmap x2 y1 ants.(n)) hex.cell_ant;
    (match hex.ant_hill with
       | None -> pixmap#set_foreground (`NAME "darkgray")
       | Some Red -> pixmap#set_foreground (`NAME "red")
       | Some Black -> pixmap#set_foreground `BLACK);
    pixmap#polygon ~filled:false l;
    if hex.food > 0 then begin
      pixmap#set_foreground `BLACK;
      let _,y3 = border.(2) in
      pixmap#string (string_of_int hex.food) ~x:(x1-5) ~y:(y3+5) ~font
    end;
  end
  
(* Initialize the pixmap before drawing. *)
let init (pixmap:GDraw.pixmap) =
  pixmap#set_foreground `WHITE;
  let width, height = pixmap#size in
  pixmap#rectangle ~x:0 ~y:0 ~width ~height ~filled:true ()

let refresh (drawing:GDraw.drawable) pixmap = 
  drawing#put_pixmap ~x:0 ~y:0 pixmap#pixmap

let draw_world drawing pixmap =
  init pixmap; 
  let s = !size in
  let m, h1, h2 = hexdim s in
  let a = m * h1 / (m + h1) in
  let hexsides = hexsides ~s ~m ~h1 ~h2 in
  for j = 0 to height - 1 do
    for i = 0 to width - 1 do
      let x, y = real_pos s m h2 i j in
      let hexsides = hexsides ~x ~y in
      draw_hex pixmap hexsides world.(i).(j);
    done
  done;
  refresh drawing pixmap

let center scroll width height ev =
  let value x size max_size =
    let mid = size /. 2. in
    let v = x -. mid in
    if v < 0. then 0.
    else 
      if x +. mid > max_size then max_size -. size
      else v
  in
  let h = scroll#hadjustment in
  let v = scroll#vadjustment in
  h#set_value (value (GdkEvent.Button.x ev) h#page_size (float width));
  v#set_value (value (GdkEvent.Button.y ev) v#page_size (float height));
  false

(* Make the area, the drawing and the pixmap containing the map. *)
let make_area scroll width height =
  let area = GMisc.drawing_area 
    ~width ~height ~packing:scroll#add_with_viewport () in
  area#misc#realize ();
  let drawing = new GDraw.drawable area#misc#window in
  let pixmap = 
    ref (GDraw.pixmap ~window:area ~width ~height ~mask:true ()) 
  in
  ignore (area#event#connect#expose 
	    ~callback:(fun _ -> refresh drawing !pixmap; false));
  area#event#add [ `BUTTON_PRESS ];
  ignore 
    (area#event#connect#button_press ~callback:(center scroll width height));
  area, drawing, pixmap

let size_world size =
  let hexsize s = 
    let h, m = hexsize s / 2, s / 2 in 
    m, m - h, m + h 
  in
  let m, h1, h2 = hexsize size in
  width * size + m, height * h2 + h1

let area, drawing, pixmap = 
  let scroll = GBin.scrolled_window 
    ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:vbox#add () 
  in
  let width, height = size_world !size in
  let area, drawing, pixmap = make_area scroll width height in
  draw_world drawing !pixmap;
  area, drawing, pixmap

let draw_cell (i, j) =
  let s = !size in
  let m, h1, h2 = hexdim s in
  let a = m * h1 / (m + h1) in
  let hexsides = hexsides ~s ~m ~h1 ~h2 in
  let x, y = real_pos s m h2 i j in
  let hexsides = hexsides ~x ~y in
  draw_hex !pixmap hexsides world.(i).(j)

let refresh () = refresh drawing !pixmap

let resize s =
  if s >= default_size / 5 && s <= default_size * 2 then begin
    size := s;
    let width, height = size_world s in
    area#set_size width height;
    pixmap := GDraw.pixmap ~window:area ~width ~height ~mask:true ();
    draw_world drawing !pixmap
  end

let resize_percent n () = resize (default_size * n / 100)
let resize_plus n () = resize (!size + n)

let start ?(init=(fun () -> ())) ~enter () =
  init ();
  ignore (window#event#connect#delete ~callback:(fun _ -> exit 0));
  window#show ();
  let enter = 
    let is_running = ref false in
    fun () -> 
      if not !is_running then ignore (Thread.create enter ());
      is_running := true
  in
  Okey.add window GdkKeysyms._Return enter;
  Okey.add window GdkKeysyms._KP_Add (resize_plus 10);
  Okey.add window GdkKeysyms._KP_Subtract (resize_plus (- 10));
  Okey.add window GdkKeysyms._KP_0 (resize_percent 100);
  Okey.add window GdkKeysyms._KP_1 (resize_percent 80);
  Okey.add window GdkKeysyms._KP_2 (resize_percent 60);
  Okey.add window GdkKeysyms._KP_3 (resize_percent 40);
  Okey.add window GdkKeysyms._KP_4 (resize_percent 20);
  Okey.add window GdkKeysyms._F1
    (fun () -> see_markers := not !see_markers; draw_world drawing !pixmap);
  GtkThread.main ()
(*  GMain.Main.main ()*)
